import os
import pathlib as pl
import uuid
from typing import TYPE_CHECKING, List, Tuple, Union

import cv2
import docx
import docx.oxml
import docx.oxml.ns
import htmldocx
import numpy as np
import PIL
import PIL.Image
import PIL.ImageDraw
import PIL.ImageFont
from docx.enum.table import WD_ALIGN_VERTICAL, WD_TABLE_ALIGNMENT
from docx.enum.text import WD_ALIGN_PARAGRAPH
from docx.oxml import OxmlElement, parse_xml
from docx.oxml.ns import nsdecls, qn
from docx.shared import Cm, Inches, Pt
from htmldocx import HtmlToDocx

from tcutility import results
from tcutility.report.formatters.generic import WordFormatter
from tcutility.report.formatters.xyz import StandardXYZFormatter
from tcutility.results import Result, read

if TYPE_CHECKING:
    import docx.document
    import docx.oxml.ns
    import docx.oxml.shared


def _add_page_numbers(document: "docx.document.Document") -> None:
    document.sections[0].footer.paragraphs[0].alignment = WD_ALIGN_PARAGRAPH.CENTER
    run = document.sections[0].footer.paragraphs[0].add_run()
    fldChar1 = OxmlElement("w:fldChar")
    fldChar1.set(qn("w:fldCharType"), "begin")

    instrText = OxmlElement("w:instrText")
    instrText.set(qn("xml:space"), "preserve")
    instrText.text = "PAGE"  # type: ignore  # text not found in docx by mypy

    fldChar2 = OxmlElement("w:fldChar")
    fldChar2.set(qn("w:fldCharType"), "end")

    run._r.append(fldChar1)
    run._r.append(instrText)
    run._r.append(fldChar2)


def _set_cell_border(cell, **kwargs):
    """
    Set a cell`s border
    Usage:

    _set_cell_border(
        cell,
        top={"sz": 12, "val": "single", "color": "#FF0000", "space": "0"},
        bottom={"sz": 12, "color": "#00FF00", "val": "single"},
        start={"sz": 24, "val": "dashed", "shadow": "true"},
        end={"sz": 12, "val": "dashed"},
    )
    """
    tc = cell._tc
    tcPr = tc.get_or_add_tcPr()

    # check for tag existnace, if none found, then create one
    tcBorders = tcPr.first_child_found_in("w:tcBorders")
    if tcBorders is None:
        tcBorders = OxmlElement("w:tcBorders")
        tcPr.append(tcBorders)

    # list over all available tags
    for edge in ("start", "top", "end", "bottom", "insideH", "insideV"):
        edge_data = kwargs.get(edge)
        if edge_data:
            tag = "w:{}".format(edge)

            # check for tag existnace, if none found, then create one
            element = tcBorders.find(qn(tag))  # type: ignore
            if element is None:
                element = OxmlElement(tag)
                tcBorders.append(element)

            # looks like order of attributes is important
            for key in ["sz", "val", "color", "space", "shadow"]:
                if key in edge_data:
                    element.set(qn("w:{}".format(key)), str(edge_data[key]))


def _set_repeat_table_header(row):
    """set repeat table row on every new page"""
    tr = row._tr
    trPr = tr.get_or_add_trPr()
    tblHeader = OxmlElement("w:tblHeader")
    tblHeader.set(qn("w:val"), "true")
    trPr.append(tblHeader)
    return row


def _set_cell_color(cell, color):
    shading_elm_1 = parse_xml(rf'<w:shd {nsdecls("w")} w:fill="{color}"/>')
    cell._tc.get_or_add_tcPr().append(shading_elm_1)


table_formatting = results.Result()
table_formatting.table.alignment = WD_TABLE_ALIGNMENT.CENTER  # type: ignore  # WD_TABLE_ALIGNMENT not found in docx by mypy
table_formatting.font.size = Pt(10.5)  # type: ignore  # Pt not found in docx by mypy


class DocxFigure:
    def __init__(self, doc: docx.document.Document, width=None, height=None):
        self.caption = ""
        self.doc = doc
        self.width = width or Cm(17.8)
        self.height = height

        self.figures = []
        self.html_parser = htmldocx.HtmlToDocx()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.write()

    def add_image(self, path: str, row: Union[int, Tuple[int]], col: Union[int, Tuple[int], None] = None, label: str = ""):
        img = cv2.imread(path)

        self.figures.append(
            {
                "path": path,
                "img": img,
                # 'col': (col, col) if isinstance(col, int) else col,
                # 'row': (row, row) if isinstance(row, int) else row,
                "col": col,
                "row": row,
                "label": label,
            }
        )

    def write(self):
        new_img = self._stitch_images()
        new_img_path = str(uuid.uuid4()) + ".png"
        cv2.imwrite(new_img_path, new_img)

        p = self.doc.add_paragraph()
        r = p.add_run()
        r.add_picture(new_img_path, width=self.width, height=self.height)

        self._write_caption()

        os.remove(new_img_path)

    def _stitch_images(self):
        grid_width, grid_height = max(fig["col"] for fig in self.figures) + 1, max(fig["row"] for fig in self.figures) + 1
        col_widths = [max(fig["img"].shape[1] for fig in self.figures if fig["col"] == index) for index in range(grid_width)]
        dpi = sum(col_widths) / (self.width / 360000 * 0.393700787)
        letter_extra_rows = int(dpi * 12 / 72) + 1

        row_heights = [max(fig["img"].shape[0] + letter_extra_rows for fig in self.figures if fig["row"] == index) for index in range(grid_height)]

        for fig in self.figures:
            height = row_heights[fig["row"]]
            width = col_widths[fig["col"]]

            fig["img"] = cv2.copyMakeBorder(fig["img"], height - fig["img"].shape[0], 0, 0, width - fig["img"].shape[1], cv2.BORDER_CONSTANT, value=[255, 255, 255])

            font_path = "Times New Roman"
            font_size = letter_extra_rows

            font = PIL.ImageFont.truetype(font_path, font_size)
            image_pil = PIL.Image.fromarray(fig["img"])  # Convert OpenCV image to PIL image
            draw = PIL.ImageDraw.Draw(image_pil)
            if fig["label"]:
                draw.text((0, 0), fig["label"], font=font, fill=(0, 0, 0))
            fig["img"] = np.array(image_pil)  # Convert PIL image back to OpenCV image

        new_img = np.zeros((sum(row_heights), sum(col_widths), 3))

        print(self.width / 360000 * 0.393700787, Inches(self.width), sum(col_widths), dpi)

        for fig in self.figures:
            start_pos = sum(row_heights[: fig["row"]]), sum(col_widths[: fig["col"]])
            end_pos = sum(row_heights[: fig["row"] + 1]), sum(col_widths[: fig["col"] + 1])
            new_img[start_pos[0] : end_pos[0], start_pos[1] : end_pos[1], :] = fig["img"]

        return new_img

    def _write_caption(self):
        paragraph = self.doc.add_paragraph("Figure S")
        paragraph.runs[-1].bold = True

        # numbering field
        run = paragraph.add_run("")
        run.bold = True

        fldChar = OxmlElement("w:fldChar")
        fldChar.set(qn("w:fldCharType"), "begin")
        run._r.append(fldChar)

        instrText = OxmlElement("w:instrText")
        instrText.text = " SEQ Figure \\* ARABIC"  # type: ignore  # text not found in docx by mypy
        run._r.append(instrText)

        fldChar = OxmlElement("w:fldChar")
        fldChar.set(qn("w:fldCharType"), "end")
        run._r.append(fldChar)

        html_doc = self.html_parser.parse_html_string(f"<b>.</b> {self.caption}")

        for add_run in html_doc.paragraphs[-1].runs:
            run = paragraph.add_run(add_run.text)
            run.bold = add_run.bold
            run.italic = add_run.italic
            run.underline = add_run.underline  # type: ignore  # underline not found in docx by mypy
            run.font.superscript = add_run.font.superscript
            run.font.subscript = add_run.font.subscript
            run.font.strike = add_run.font.strike
            run.font.size = add_run.font.size


class DocxTable:
    def __init__(self, file: Union[str, pl.Path, docx.document.Document] = "test.docx"):
        self.file = file
        self.dont_save = False
        if isinstance(file, docx.document.Document):
            self.dont_save = True
            self.doc = file
        else:
            if not os.path.exists(file):
                self.doc = docx.Document()
            else:
                self.doc = docx.Document(str(file))

        self.caption = ""
        self.columns = []
        self.column_options = []
        self.footnote = None
        self.rows = []
        self.mergers = []
        self.html_parser = htmldocx.HtmlToDocx()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.write()

    def add_column(self, name, **kwargs):
        self.columns.append(["single", name, kwargs])
        self.column_options.append(kwargs)

    def add_column_group(self, group_name, column_names, **kwargs):
        self.columns.append(["grouped", group_name, column_names])
        [self.column_options.append(kwargs) for _ in column_names]

    def add_row(self, data):
        self.rows.append(["data", data])

    def add_header_row(self, name):
        self.rows.append(["header", name])

    def add_empty_row(self):
        self.rows.append(["empty"])

    def merge_cells(self, x, y):
        if isinstance(x, int):
            x = (x, x)
        if isinstance(y, int):
            y = (y, y)

        self.mergers.append([x, y])

    def _correct_table_dims(self):
        num_cols = 2 * len([col for col in self.columns if col[0] == "single"]) + sum([len(col[2]) + 1 for col in self.columns if col[0] == "grouped"]) - 1
        num_rows = 2 + len(self.rows)

        for _ in range(num_cols - len(self.tab.columns)):
            self.tab.add_column(Cm(0.1))
        for _ in range(num_rows - len(self.tab.rows)):
            self.tab.add_row()

    @property
    def shape(self):
        Ncols = 0

        for col in self.columns:
            if col[0] == "single":
                Ncols += 2
            if col[0] == "grouped":
                Ncols += len(col[2]) + 1

        return 2 + len(self.rows), Ncols - 1

    def write(self):
        self._write_caption()
        # write the table caption
        self.doc.paragraphs[-1].alignment = WD_ALIGN_PARAGRAPH.JUSTIFY

        self.tab = self.doc.add_table(1, 1)
        self._correct_table_dims()

        for x, y in self.mergers:
            self.tab.cell(x[0], y[0]).merge(self.tab.cell(x[1], y[1]))

        num_cols = 2 * len([col for col in self.columns if col[0] == "single"]) + sum([len(col[2]) + 1 for col in self.columns if col[0] == "grouped"]) - 1
        # create a table
        self.tab.alignment = table_formatting.table.alignment  # type: ignore # Results object does not have typing

        # write the column headers
        spacing_columns = []
        col_idx = 0
        for col in self.columns:
            if col[0] == "single":
                self.write_cell(1, col_idx, col[1], bold=True, font_size=table_formatting.font.size)  # type: ignore  # results object does not have typing
                spacing_columns.append(col_idx + 1)
                col_idx += 2

            if col[0] == "grouped":
                _ = self.write_cell(0, (col_idx, col_idx + len(col[2]) - 1), col[1], bold=True, bottom={"sz": 12, "val": "single", "color": "#000000"}, font_size=table_formatting.font.size)  # type: ignore  # results object does not have typing

                for i, val in enumerate(col[2]):
                    self.write_cell(1, col_idx + i, val, bold=True, font_size=table_formatting.font.size)  # type: ignore  # results object does not have typing

                spacing_columns.append(col_idx + len(col[2]))
                col_idx += len(col[2]) + 1

        # set the lines for the top and bottom header rows
        for i in range(num_cols):
            _set_cell_border(self.tab.cell(0, i), top={"sz": 12, "val": "single", "color": "#000000"})
            _set_cell_border(self.tab.cell(1, i), bottom={"sz": 12, "val": "single", "color": "#000000"})

        for j, row in enumerate(self.rows):
            if row[0] == "data":
                for i in range(num_cols):
                    if i in spacing_columns:
                        continue
                    num_spacing_past = len([k for k in spacing_columns if (k - 1) < i])
                    self.write_cell(j + 2, i, row[1][i - num_spacing_past], font_size=table_formatting.font.size, **self.column_options[i - num_spacing_past])  # type: ignore  # results object does not have typing

            if row[0] == "header":
                _ = self.write_cell(
                    j + 2,
                    (0, num_cols - 1),
                    row[1],
                    bold=True,
                    top={"sz": 12, "val": "single", "color": "#000000"},
                    bottom={"sz": 12, "val": "single", "color": "#000000"},
                    bkgr_color="F2F2F2",
                    font_size=table_formatting.font.size,  # type: ignore  # font not found in docx by mypy
                )

            if row[0] == "empty":
                for i in range(num_cols):
                    self.write_cell(j + 2, i, "")

        _set_repeat_table_header(self.tab.rows[0])
        _set_repeat_table_header(self.tab.rows[1])

        self._write_footnote()

        if not self.dont_save:
            self.doc.save(str(self.file))

    def write_cell(self, row, col, text, alignment="center", vert_alignment="center", bold=None, italic=None, bkgr_color=None, font_size=None, **kwargs):
        if isinstance(row, int) and isinstance(col, int):
            cell = self.tab.cell(row, col)
        else:
            if isinstance(row, int):
                row = (row, row)
            if isinstance(col, int):
                col = (col, col)
            cell = self.tab.cell(row[0], col[0]).merge(self.tab.cell(row[1], col[1]))

        if len(cell.paragraphs[0].runs) > 0:
            return

        # cell.text = str(text).strip()
        alignment = {
            "center": WD_ALIGN_PARAGRAPH.CENTER,
            "left": WD_ALIGN_PARAGRAPH.LEFT,
            "right": WD_ALIGN_PARAGRAPH.RIGHT,
        }[alignment]

        vert_alignment = {
            "center": WD_ALIGN_VERTICAL.CENTER,
            "top": WD_ALIGN_VERTICAL.TOP,
            "bottom": WD_ALIGN_VERTICAL.BOTTOM,
        }[vert_alignment]

        # for txt, settings in parse_text(text):
        self.html_parser.add_html_to_cell(text.replace("-", "–"), cell)

        cell.paragraphs[0].alignment = alignment
        cell.vertical_alignment = vert_alignment
        for run in cell.paragraphs[0].runs:
            if bold:
                run.bold = bold
            if italic:
                run.italic = italic
            if font_size:
                run.font.size = font_size

        _set_cell_border(cell, **kwargs)
        if bkgr_color is not None:
            _set_cell_color(cell, bkgr_color)

        return cell

    def add_footnote(self, text, bold=None, italic=None, font_size=None):
        self.footnote = (text, bold, italic, font_size)

    def _write_footnote(self):
        if self.footnote is None:
            return

        # set the lower line
        for col in range(self.shape[1]):
            cell = self.tab.cell(self.shape[0] - 1, col)
            _set_cell_border(cell, bottom={"sz": 12, "val": "single", "color": "#000000"})

        self.html_parser.add_html_to_document(self.footnote[0], self.doc)

        for run in self.doc.paragraphs[0].runs:
            if self.footnote[1]:
                run.bold = self.footnote[1]
            if self.footnote[2]:
                run.italic = self.footnote[2]
            if self.footnote[3]:
                run.font.size = self.footnote[3]

    def _write_caption(self):
        paragraph = self.doc.add_paragraph("Table S")
        paragraph.runs[-1].bold = True

        # numbering field
        run = paragraph.add_run("")
        run.bold = True

        fldChar = OxmlElement("w:fldChar")
        fldChar.set(qn("w:fldCharType"), "begin")
        run._r.append(fldChar)

        instrText = OxmlElement("w:instrText")
        instrText.text = " SEQ Table \\* ARABIC"  # type: ignore  # text not found in docx by mypy
        run._r.append(instrText)

        fldChar = OxmlElement("w:fldChar")
        fldChar.set(qn("w:fldCharType"), "end")
        run._r.append(fldChar)

        html_doc = self.html_parser.parse_html_string(f"<b>.</b> {self.caption}")

        for add_run in html_doc.paragraphs[-1].runs:
            run = paragraph.add_run(add_run.text)
            run.bold = add_run.bold
            run.italic = add_run.italic
            run.underline = add_run.underline  # type: ignore  # underline not found in docx by mypy
            run.font.superscript = add_run.font.superscript
            run.font.subscript = add_run.font.subscript
            run.font.strike = add_run.font.strike
            run.font.size = add_run.font.size


class SI:
    def __init__(self, path: Union[str, pl.Path], overwrite: bool = False, font: str = "Arial") -> None:
        """Initializes the SI class for creating supporting information (SI) files in Microsoft Word format.

        This class is responsible for creating and managing a Microsoft Word document that serves as supporting information (SI) for reports or publications.
        It allows for the addition of various elements such as text, headings, and formatted content from HTML.

        Args:
            path (str | pl.Path): The location of the Word file. Does not have to have a file-extension.
            overwrite (bool, optional): Whether to append to or overwrite the file. Defaults to False.
            font (str, optional): The font to be used in the document. Defaults to "Arial".

        Attributes:
            path (pl.Path): The path to the Word document.
            doc (docx.Document): The Word document object.
        """
        self.path = pl.Path(path).with_suffix(".docx")
        self.doc = docx.Document()

        if not os.path.exists(self.path) or overwrite:
            self.doc = docx.Document()
        else:
            self.doc = docx.Document(str(self.path))

        self.doc.styles["Normal"].font.name = "Times New Roman"  # type: ignore  # font not found in docx by mypy
        self.doc.styles["Normal"].font.size = Pt(12)  # type: ignore  # font not found in docx by mypy
        self.doc.styles["Normal"].paragraph_format.space_after = 1  # type: ignore  # paragraph_format not found in docx by mypy
        self.html_parser = htmldocx.HtmlToDocx()  # type: ignore  # HtmlToDocx not found in docx by mypy

        # Set the font to the specified font
        self.doc.styles["Normal"].font.name = font  # type: ignore  # font not found in docx by mypy

    def __enter__(self):
        """Enables the use of the class as a context manager."""
        return self

    def __exit__(self, *args):
        for section in self.doc.sections:
            # 1.9 cm
            section.left_margin = Cm(1.9)
            section.right_margin = Cm(1.9)

        _add_page_numbers(self.doc)

        self.doc.save(str(self.path))

    def add_figure(self, *args, **kwargs) -> DocxFigure:
        return DocxFigure(self.doc, *args, **kwargs)

    def add_table(self) -> DocxTable:
        return DocxTable(self.doc)

    def add_xyz(self, obj: Union[str, Result, pl.Path], title: Union[str, None] = None, formatter: WordFormatter = StandardXYZFormatter()) -> None:
        """Adds XYZ formatted content to the document.

        This method is responsible for adding the coordinates and information about a calculation to the supporting information document.
        It includes details such as the electronic bond energy, Gibb's free energy, enthalpy, imaginary mode, and the coordinates of the molecule.

        Args:
            obj: A string specifying a calculation directory or a `TCutility.results.Result` object from a calculation.
            title: The title to be written before the coordinates and information. If None, no title is added.
            formatter: The formatter to be used for formatting the content. Defaults to `StandardXYZFormatter`.

        Returns:
            None
        """
        ret_str = ""

        # Add the formatted content to the document
        if isinstance(obj, str) or isinstance(obj, pl.Path):
            calc_path = pl.Path(obj)
            if not calc_path.is_dir():
                return  # raise ValueError(f"Invalid calculation directory: {calc_path}")
            obj = read(obj)

        ret_str += formatter.format(obj, title=title)

        # print(ret_str)
        parser = HtmlToDocx()
        parser.add_html_to_document(ret_str, self.doc)
        return

    def add_heading(self, text: str, level: int = 1) -> None:
        """Adds a heading to the document.

        This method allows for the addition of a heading to the Word document, with customizable text and level.

        Args:
            text (str): The text of the heading.
            level (int, optional): The level of the heading (determines the size and style). Defaults to 1.
        """
        self.doc.add_heading(text, level)

    def add_page_break(self) -> None:
        self.doc.add_page_break()

    def add_toc(self) -> None:
        title = self.doc.add_paragraph()
        title_run = title.add_run("Contents")
        title_run.bold = True
        title_run.font.size = Pt(14)
        title.paragraph_format.space_after = Pt(10)

        # figures
        paragraph = self.doc.add_paragraph()
        paragraph.paragraph_format.space_after = Pt(10)
        run = paragraph.add_run()
        fldChar = OxmlElement("w:fldChar")
        fldChar.set(qn("w:fldCharType"), "begin")
        fldChar.set(qn("w:dirty"), "true")
        instrText = OxmlElement("w:instrText")
        instrText.set(qn("xml:space"), "preserve")
        instrText.text = 'TOC \\h \\z \\c "Figure"'  # type: ignore # "Table" of list of table and "Figure" for list of figure
        fldChar2 = OxmlElement("w:fldChar")
        fldChar2.set(qn("w:fldCharType"), "separate")
        fldChar3 = OxmlElement("w:t")
        fldChar3.text = "Right-click to update field."  # type: ignore # text not found in docx by mypy
        fldChar2.append(fldChar3)

        fldChar4 = OxmlElement("w:fldChar")
        fldChar4.set(qn("w:fldCharType"), "end")

        run._r.append(fldChar)
        run._r.append(instrText)
        run._r.append(fldChar2)
        run._r.append(fldChar4)

        # tables
        paragraph = self.doc.add_paragraph()
        paragraph.paragraph_format.space_after = Pt(10)
        run = paragraph.add_run()
        fldChar = OxmlElement("w:fldChar")
        fldChar.set(qn("w:fldCharType"), "begin")
        fldChar.set(qn("w:dirty"), "true")
        instrText = OxmlElement("w:instrText")
        instrText.set(qn("xml:space"), "preserve")
        instrText.text = 'TOC \\h \\z \\c "Table"'  # type: ignore # "Table" of list of table and "Figure" for list of figure
        fldChar2 = OxmlElement("w:fldChar")
        fldChar2.set(qn("w:fldCharType"), "separate")
        fldChar3 = OxmlElement("w:t")
        fldChar3.text = "Right-click to update field."  # type: ignore # text not found in docx by mypy
        fldChar2.append(fldChar3)

        fldChar4 = OxmlElement("w:fldChar")
        fldChar4.set(qn("w:fldCharType"), "end")

        run._r.append(fldChar)
        run._r.append(instrText)
        run._r.append(fldChar2)
        run._r.append(fldChar4)


def get_subdirs(root_folder: pl.Path) -> List[pl.Path]:
    """Iteratively searches through a folder and returns all the most nested subdirs."""
    most_nested_subdirs = []
    for root, dirs, files in os.walk(root_folder):
        # If 'dirs' is empty, it means 'root' contains no subdirectories, thus it is most nested.
        if not dirs:
            most_nested_subdirs.append(pl.Path(root))
    return most_nested_subdirs


def replace_files_rkf_to_ams_rkf(root_folder: pl.Path) -> None:
    """Iteratively searches through a folder and replaces all files with the extension '.rkf' to '.ams.rkf', except if the file has 'adf.rkf' in the name."""
    for root, dirs, files in os.walk(root_folder):
        for file in files:
            if file.endswith(".rkf") and "adf.rkf" not in file:
                new_name = file.replace(".rkf", ".ams.rkf")
                os.rename(pl.Path(root) / file, pl.Path(root) / new_name)


def main():
    from tcutility.results import read

    calc_dir = pl.Path("__file__").resolve().parents[0] / "test" / "fixtures"
    main_path = pl.Path("__file__").resolve().parents[0] / "examples"

    # all_subdirs = [folder for folder in calc_dir.iterdir() if folder.is_dir()]
    all_subdirs = get_subdirs(calc_dir)
    res_objects = []
    for dir_ in all_subdirs:
        try:  # Try to read the results of the calculation
            res_objects.append(read(dir_))
        except Exception:
            pass

    with SI(main_path / "test", overwrite=False) as si:
        si.add_heading("SI project")
        for obj in res_objects:
            si.add_xyz(obj=obj)


if __name__ == "__main__":
    main()



"""Module containing functions for calculating geometrical parameters for molecules"""

from typing import Dict, List, Tuple

import click
from strenum import StrEnum
from tcutility import geometry, molecule, results


class GeometricParameter(StrEnum):
    """Enum class for geometric parameters"""

    COORDINATE = "Coordinate"
    DISTANCE = "Distance"
    ANGLE = "Angle"
    DIHEDRAL = "Dihedral"
    PYRAMIDAL = "Pyramidal"
    SUMOFANGLES = 'SumOfAngles'


geometric_character_to_info_mapping: Dict[GeometricParameter, Tuple[str, str, int]] = {
    GeometricParameter.COORDINATE: ("Coordinate", "Å", 6),
    GeometricParameter.DISTANCE: ("Distance", " Å", 3),
    GeometricParameter.ANGLE: ("Angle", "°", 2),
    GeometricParameter.DIHEDRAL: ("Dihedral", "°", 2),
    GeometricParameter.PYRAMIDAL: ("Pyramidal", "°", 2),
    GeometricParameter.SUMOFANGLES: ("SumOfAngles", "°", 2),
}


@click.command()
@click.argument("path", type=click.Path(exists=True))
@click.argument("atom_indices", type=str, nargs=-1)
@click.option("-p", "--pyramidal", is_flag=True, default=False, help="Instead of calculating a dihedral angle, calculate pyramidalisation angle.")
@click.option("-s", "--soa", is_flag=True, default=False, help="Instead of calculating a dihedral angle, calculate sum-of-angles angle.")
def calculate_geometry_parameter(path: str, atom_indices: List[str], pyramidal: bool, soa: bool) -> None:
    """
    \b
    Calculate geometrical parameters for atoms at the provided ATOM_INDICES (NB: counting starts at 1).
    PATH should be an XYZ-file or a calculation directory.
    
    \b
    For 0 indices we return all coordinates of the molecule in PATH.
    For 1 index this program returns the cartesian coordinate for this atom.
    For 2 indices return bond length between atoms.
    For 3 indices return bond angle, with the second index being the central atom.
    For 4 indices return dihedral angle by calculating the angle between normal vectors
        described by atoms at indices 1-2-3 and indices 2-3-4.
    \b
    If the -p/--pyramidal flag is turned on it calculates 360° - ang1 - ang2 - ang3, 
        where ang1, ang2 and ang3 are the angles described by indices 2-1-3, 3-1-4 
        and 4-1-2 respectively.
    If the -s/--soa flag is turned on it calculates ang1 + ang2 + ang3.
    """
    try:
        mol = molecule.load(path)
    except IsADirectoryError:
        mol = results.read(path).molecule.output

    assert 0 <= len(atom_indices) <= 4, f"Number of atom indices must be 1, 2, 3 or 4, not {len(atom_indices)}"

    if len(atom_indices) == 0:
        mol.delete_all_bonds()
        print(mol)
        return

    atom_indices = [int(i) - 1 for i in atom_indices]

    atoms = "-".join([f"{mol[i+1].symbol}{i+1}" for i in atom_indices])

    param_value = geometry.parameter(mol, *atom_indices, pyramidal=pyramidal, sum_of_angles=soa)

    if len(atom_indices) == 1:
        param_type, unit, precision = geometric_character_to_info_mapping[GeometricParameter.COORDINATE]
        param_value = "  ".join([f"{x: .{precision}f}" for x in mol[atom_indices[0] + 1].coords])
        print(f"{param_type}({atoms}) = {param_value} {unit}")
        return

    elif len(atom_indices) == 2:
        param_type = GeometricParameter.DISTANCE
    elif len(atom_indices) == 3:
        param_type = GeometricParameter.ANGLE
    elif len(atom_indices) == 4:
        if pyramidal:
            param_type = GeometricParameter.PYRAMIDAL
        elif soa:
            param_type = GeometricParameter.SUMOFANGLES
        else:
            param_type = GeometricParameter.DIHEDRAL
    else:
        raise ValueError("Invalid number of atom indices")

    param_type, unit, precision = geometric_character_to_info_mapping[param_type]
    print(f"{param_type}({atoms}) = {param_value: .{precision}f}{unit}")


if __name__ == "__main__":
    calculate_geometry_parameter()



import paramiko
import os
from tcutility import log, results, cache
from datetime import datetime
import subprocess as sp
import shutil
import atexit
import uuid


_open_connections = []

class Connection:
    '''
    Main class used for creating and using SSH sessions to remote servers.
    It gives you the option to execute any shell code, but also provides useful default commands (for example, ``cd`` and ``ls``). 
    The ``Connection`` class also allows you to download and upload files between your local machine and the remote.

    Args:
        server: the adress of the server you want to connect to. You can prepend the server adress with your username separated from the adress with a ``@`` character.
            For example: ``Connection('username@server.address.nl')`` is the same as ``Connection('server.address.nl', 'username')``.
        username: the username used to log in to the remote server.
        key_filename: if you cannot log in using only the ``ssh`` command you can try to give the 
            filename of the private key that matches a public key on the server.

    Usage:
        This class is a context manager and the ``with``-syntax should be used to open and automatically close connections.
        For example, to open a connection to the Snellius supercomputer we use the following code:

        .. code-block::
        
            from tcutility.connect import Connection
            
            with Connection('username@server.address.nl') as server:
                print(server.pwd())  # this will print the home-directory of the logged-in user
                server.cd('example/path/to/some/data')
                print(server.pwd())  # ~/example/path/to/some/data

    .. warning::
        Currently we only support logging in using SSH keys. Make sure that you can log in to the remote with SSH keys. 
        There might be server specific instructions on how to enable this authentication method.
    '''
    def __init__(self, server: str = None, username: str = None, key_filename: str = None):
        if server is not None:
            self.server = server

        if username is not None:
            self.username = username

        if server is not None and '@' in server:
            self.username, self.server = server.split('@')

        self.key_filename = key_filename
        if isinstance(key_filename, str):
            self.key_filename = os.path.abspath(os.path.expanduser(key_filename))

    def open(self):
        self.__enter__()

    def __enter__(self):
        log.debug(f'{self}: opening connection ...')
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.client.connect(self.server, username=self.username, key_filename=self.key_filename)
        log.debug(f'{self}: connection opened!')
        # store the home directory so we can use it later to get absolute paths
        self.home = self.client.exec_command('pwd')[1].read().decode().strip()
        self.currdir = self.home
        _open_connections.append(self)
        return self

    def close(self):
        self.__exit__()

    def __exit__(self, *args, **kwargs):
        _open_connections.remove(self)
        self.client.close()
        log.debug(f'{self}: connection closed.')

    def __repr__(self):
        return f'Connection({self.username}@{self.server})'

    def full_path(self, path: str) -> str:
        if path.startswith('/'):
            return path
        return os.path.join(self.currdir, path)

    def execute(self, command: str) -> str:
        '''
        Run a command on the server and return the output.

        Args:
            command: the command to run on the server.

        Returns:
            Data written in ``stdout`` after the command was run.

        Raises:
            ``RuntimeError`` with error message if there was something printed to the ``stderr``.

        .. note::
            The ``__call__`` method redirects to this method. This means you can directly call the ``Connection`` object with your command.
        '''
        log.debug(f'{self}[{self.currdir}]: {command}')
        command =  f'cd {self.currdir}; {command}'
        _, stdout, stderr = self.client.exec_command(command)
        stdout = stdout.read().decode()
        stderr = stderr.read().decode()
        if stderr:
            raise RuntimeError(stderr)

        return stdout.strip()

    def __call__(self, *args, **kwargs):
        return self.execute(*args, **kwargs)

    def ls(self, path='') -> results.Result:
        '''
        Run the ``ls`` program and return useful information about the paths.

        Args:
            path: the path to run ls on.

        Returns:
            :Result object containing information from the output of the ``ls`` program. 
                The keys are the path names and the values contain the information.

                - **owner (str)** - the username of the owner of the path.
                - **date (datetime.datetime)** - ``datetime`` object holding the date the file was created.
                - **is_dir (bool)** - whether the path is a directory.
                - **is_hidden (bool)** - whether the path is hidden.
                - **permissions (str)** - the permissions given to the path.
        '''
        out = self.execute(f'TZ="UTC" ls -lAFp --full-time {path}')
        lines = out.splitlines()
        ret = results.Result()
        for line in lines[1:]:  # first line is a total count
            parts = line.split()
            name = parts[-1]
            is_dir = False
            if name.endswith('/'):
                is_dir = True
                name = name[:-1]

            ret[name].is_dir = is_dir

            timestamp = " ".join(parts[5:7])
            ret[name].date = datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S.%f000')
            ret[name].owner = parts[2]
            ret[name].is_hidden = name.startswith('.')
            ret[name].permissions = parts[0]

        return ret

    def cd(self, path='~'):
        '''
        Run the ``cd`` command. 

        Args:
            path: the path to change directories to. This is relative to the current directory.

        .. note::
            Due to limitations with some servers (e.g. Snellius) we do not actually run the ``cd`` command, 
            but update the internal ``Connection.currdir`` attribute. Before running any command we prepend with ``cd {self.currdir}; ...``.
            In this way we run commands from the correct directory.
        '''
        log.debug(f'{self}[{self.currdir}]: cd {path}')
        path = path.replace('~', self.home)
        if path.startswith('/'):
            self.currdir = path
        else:
            self.currdir = os.path.normpath(os.path.join(self.currdir, path))

    def pwd(self) -> str:
        '''
        Run the ``pwd`` command.

        .. note::
            Due to limitations with some servers (e.g. Snellius) we do not actually run the ``pwd`` command, 
            instead we return the internal ``Connection.currdir`` attribute. See the :func:`Connection.cd` method for more details.
        '''
        log.debug(f'{self}[{self.currdir}]: pwd')
        return self.currdir

    def mkdir(self, dirname):
        '''
        Run the ``mkdir`` command.

        Args:
            dirname: the name of the directory to make. This is relative to the current directory.
        '''
        self.execute(f'mkdir -p {dirname}')

    def rm(self, file_path):
        self.execute(f'rm {file_path}')

    def rmtree(self, dirname):
        self.execute(f'rm -r {dirname}')

    def download(self, server_path: str, local_path: str):
        '''
        Download a file from the server and store it on your local machine.

        Args:
            server_path: the path on the server to the file to download. The path is relative to the current directory.
            local_path: the path on the local machine where the file is stored.
        '''
        server_path = os.path.normpath(self.full_path(server_path))

        log.debug(f'{self}: download {server_path} {local_path}')
        with self.client.open_sftp() as sftp:
            sftp.get(server_path, local_path)
        log.debug(f'{self}: download completed!')

    def upload(self, local_path: str, server_path: str = None):
        '''
        Upload a file from your local machine to the server. 
        If the ``server_path`` is not given, store it in the current directory.

        Args:
            local_path: the path on the local machine where the file to be uploaded is stored.
            server_path: the path to upload the file to. 
                If not given or set to ``None`` we upload the file to the current directory with the same filename.
        '''
        if server_path is None:
            server_path = os.path.join(self.currdir, os.path.basename(local_path))
        elif not server_path.startswith('/'):
            server_path = os.path.join(self.currdir, server_path)
        server_path = os.path.normpath(server_path)

        log.debug(f'{self}: upload {local_path} {server_path}')
        with self.client.open_sftp() as sftp:
            sftp.put(local_path, server_path)
        log.debug(f'{self}: upload completed!')

    def path_exists(self, path: str):
        test = self.execute(f'test -e {path}; echo $?')
        return test == "0"

    def open_file(self, file_path: str):
        return ServerFile(file_path, self)

    def chmod(self, rights: int, file_path: str):
        self.execute(f'chmod {rights} {file_path}')


class ServerFile:
    def __init__(self, file_path: str, server: Connection):
        self.server = server
        self.file_path = file_path
        self.tmp_path = str(uuid.uuid4())
        self.file = open(self.tmp_path, mode='w+')

    def __enter__(self):
        return self.file

    def __exit__(self, *args):
        self.file.close()
        self.server.upload(self.tmp_path, self.file_path)
        os.remove(self.tmp_path)



class Server(Connection):
    '''
    Helper subclass of :class:``Connection`` that is used to quickly connect to a specific server.
    The constructor takes only the username as the server url is already set. You can also specify 
    default settings for ``sbatch`` calls, for example the partition or time-limits.
    '''
    server = None
    sbatch_defaults = {}
    preamble_defaults = {}
    postamble_defaults = {}
    program_modules = {}

    def __init__(self, username=None):
        super().__init__(self.server, username=username)

    def __repr__(self):
        return f'{type(self).__name__}({self.username})'


class Local:
    server = None
    sbatch_defaults = {}
    preamble_defaults = {}
    postamble_defaults = {}
    program_modules = {}

    def __init__(self):
        self.currdir = os.getcwd()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        ...

    def execute(self, command: str) -> str:
        """
        Execute a command on the local machine and return the output.
    
        Args:
            command: the command to run.

        .. note::
            We use ``subprocess.check_output`` with the ``shell=True`` argument enabled.
        """
        try:
            output = sp.check_output(command, shell=True).decode()
            return output
        except sp.CalledProcessError:
            print('COMMAND: ', command)
            raise

    def mkdir(self, dirname):
        os.makedirs(os.path.join(self.currdir, dirname), exist_ok=True)

    def rm(self, file_path):
        os.remove(file_path) if os.path.exists(file_path) else ...

    def rmtree(self, dirname):
        shutil.rmtree(dirname) if os.path.exists(dirname) else ...

    def download(self, server_path: str, local_path: str):
        shutil.copy2(os.path.join(self.currdir, server_path), os.path.join(self.currdir, local_path))

    def upload(self, local_path: str, server_path: str = None):
        shutil.copy2(os.path.join(self.currdir, local_path), os.path.join(self.currdir, server_path))

    def path_exists(self, path: str) -> bool:
        return os.path.exists(os.path.join(self.currdir, path))

    def open_file(self, file_path: str):
        return open(file_path, mode='w+')

    def chmod(self, rights: int, path: str):
        os.chmod(os.path.join(self.currdir, path), int(str(rights), base=8))

    def pwd(self):
        return os.getcwd()

    def ls(self, dirname):
        return os.listdir(dirname)



class Ada(Server):
    '''
    Default set-up for a connection to the Ada cluster. By default we use the ``tc`` partition.
    '''
    server = 'ada.labs.vu.nl'
    sbatch_defaults = {
        'p': 'tc',
        'ntasks_per_node': 16,
        'N': 1,
        'mem': 250000,
    }
    preamble_defaults = {
        'AMS': [
            'export SCM_TMPDIR="/scratch/$SLURM_JOBID"',
            "srun mkdir -p $SCM_TMPDIR",
            "chmod 700 $SCM_TMPDIR",
        ]
    }
    program_modules = {
        'AMS': {
            '2021': 'module load shared ams/2021.102',
            '2022': 'module load shared ams/2022.103',
            '2023': 'module load shared ams/2023.101',
            '2024': 'module load shared ams/2024.102',
            'latest': 'module load shared ams/2024.102',
        }
    }
    postamble_defaults = {
        'AMS': [
            "srun rm -rf $SCM_TMPDIR",
        ]
    }


class Snellius(Server):
    '''
    Default set-up for a connection to the Snellius cluster. By default we use the ``rome`` partition and a time-limit set to ``120:00:00``.
    '''
    server = 'snellius.surf.nl'
    sbatch_defaults = {
        'p': 'rome',
        't': '120:00:00',
        'ntasks_per_node': 16,
        'N': 1,
    }
    program_modules = {
        'AMS': {
            '2023': 'module load 2023 AMS/2023.104-intelmpi',
            '2024': 'module load 2024 AMS/2024.104-intelmpi-aocl',
            'latest': 'module load 2024 AMS/2024.104-intelmpi-aocl',
        }
    }


@cache.cache
def get_current_server() -> Server:
    '''
    Return the `Server`-subclass of the server location of the current shell.
    If the server location could not be detected returns `Local`.
    '''
    ifconfig = sp.check_output('ifconfig')
    parts = ifconfig.decode().split()
    adresses = []
    for i, part in enumerate(parts):
        if part == 'inet':
            adresses.append(parts[i+1].split('.')[:3])

    # print(ifconfig)
    for cls in Server.__subclasses__():
        ping = sp.check_output(['ping', cls.server, '-c', '1'])
        for part in ping.decode().split():
            if part.startswith('(') and part.endswith('):'):
                ip_address = part[1:-2].split('.')[:3]
                if ip_address in adresses:
                    return cls
    return Local


def _close_all_connections():
    for conn in _open_connections:
        conn.__exit__()

atexit.register(_close_all_connections)

if __name__ == '__main__':
    print(get_current_server())
